(*----------------------------------------------------------------------------*
 *  Direct3D sample for DirectX 8.1 demostrating how to create and render     *
 *  Dot product (Dot3) bumpmapped / modulated font. Font normal map generated *
 *  with D3DX8.1 function D3DXComputeNormalMap which itself requires height   *
 *  map. So height map is generated from font bitmap by averaging intensivity *
 *  of 3x3 square of pixels.                                                  *
 *                                                                            *
 *  Note: This sample requires DirectX 8.1 to compile and run!!!              *
 *                                                                            *
 *  Copyright (c) 2002 by Alexey Barkovoy                                     *
 *  E-Mail: clootie@reactor.ru                                                *
 *                                                                            *
 *  Modified: 29-May-2002                                                     *
 *                                                                            *
 *  Latest version can be downloaded from:                                    *
 *     http://clootie.narod.ru                                                *
 *----------------------------------------------------------------------------*)

unit Text3DUnit;

interface

uses
  Windows, Messages, SysUtils,
   Direct3D8 ,
  D3DX8, D3DApp, D3DFont, D3DUtil, DXUtil, CommDlg, D3DRes, D3DFontDot3;

// C++Builder compile hack
{$NOINCLUDE D3DFontDot3}
{$HPPEMIT '#include "D3DFontDot3.hpp"'}

type
//-----------------------------------------------------------------------------
// Name: class CMyD3DApplication
// Desc: Application class. The base class (CD3DApplication) provides the
//       generic functionality needed in all Direct3D samples. CMyD3DApplication
//       adds functionality specific to this sample program.
//-----------------------------------------------------------------------------
  CMyD3DApplication = class(CD3DApplication)
    m_pFont: CD3DFont;
    m_pDot3Font: TD3DFontDot3;
    m_pStatsFont: CD3DFont;

    m_strFont: array[0..99] of Char;
    m_dwFontSize: DWORD;

    MouseNormal, CircularNormal: TD3DXVector3;

    Hue: Byte;
  protected
    function OneTimeSceneInit: HRESULT; override;
    function InitDeviceObjects: HRESULT; override;
    function RestoreDeviceObjects: HRESULT; override;
    function InvalidateDeviceObjects: HRESULT; override;
    function DeleteDeviceObjects: HRESULT; override;
    function Render: HRESULT; override;
    function FrameMove: HRESULT; override;
    function FinalCleanup: HRESULT; override;
  public
    constructor Create;
    function MsgProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM; lParam: LPARAM): LRESULT; override;
    procedure ChangeFont;
  end;

implementation

// Additional routines
type
  TRGBRec = packed record
    B, G, R, a: Byte;
  end;

  THSBRec = packed record
    H, S, B, a: Byte;
  end;

type
  TToneArray = array [0..255] of TRGBRec;

function HSB(H, S, B: Byte): THSBRec;
begin
  Result.H:= H;
  Result.S:= S;
  Result.B:= B;
  Result.a:= 0;
end;

procedure DivMod(Dividend: Integer; Divisor: Word;
  var Result, Remainder: Word);
asm
        PUSH    EBX
        MOV     EBX,EDX
        MOV     EDX,EAX
        SHR     EDX,16
        DIV     BX
        MOV     EBX,Remainder
        MOV     [ECX],AX
        MOV     [EBX],DX
        POP     EBX
end;

function HSBtoRGB(HSB: THSBRec): TRGBRec;
 procedure DoLuminescency(var V: Byte; L: Byte);
 begin
   if L > 120
   then V:= V + ((L - 120)*(255 - V)) div 120
   else V:= V - ((120 - L)*V) div 120;
 end;
var
  ColInd, Rest: Word;
begin
  Assert(not ((HSB.H > 240) or (HSB.S > 240) or (HSB.B > 240)),
    'HSB value out of range');

  // calculate Hue
  DivMod(HSB.H, 40, ColInd, Rest);
  case ColInd of
    0: begin //  = 255,      // > 0
       Result.R:= 255;
       Result.G:= (255*Rest) div 40;
       Result.B:= 0;
     end;
    1: begin //  ,  = 255      // > 40
       Result.R:= (255*(40 - Rest)) div 40;
       Result.G:= 255;
       Result.B:= 0;
     end;
    2: begin //  = 255,        // > 80
       Result.R:= 0;
       Result.G:= 255;
       Result.B:= (255*Rest) div 40;
     end;
    3: begin //  ,  = 255        // > 120
       Result.R:= 0;
       Result.G:= (255*(40 - Rest)) div 40;
       Result.B:= 255;
     end;
    4: begin //  = 255,        // > 160
       Result.R:= (255*Rest) div 40;
       Result.G:= 0;
       Result.B:= 255;
     end;
    5: begin //  ,  = 255        // > 200
       Result.R:= 255;
       Result.G:= 0;
       Result.B:= (255*(40 - Rest)) div 40;
     end;
  end;

  // add Saturation
  Result.R:= Result.R + ((240 - HSB.S)*(127 - Result.R)) div 240;
  Result.G:= Result.G + ((240 - HSB.S)*(127 - Result.G)) div 240;
  Result.B:= Result.B + ((240 - HSB.S)*(127 - Result.B)) div 240;

  // add Luminescency
  DoLuminescency(Result.R, HSB.B);
  DoLuminescency(Result.G, HSB.B);
  DoLuminescency(Result.B, HSB.B);
end;

// function compares 2 HSB records to verify equality
function IsEqualHSB(HSB1, HSB2: THSBRec): Boolean;
begin
  Result:= (HSB1.H = HSB2.H) and (HSB1.S = HSB2.S) and (HSB1.B = HSB2.B);
end;

///////////////////////////////////////////////////////////
// CMyD3DApplication.
///////////////////////////////////////////////////////////

constructor CMyD3DApplication.Create;
begin
  inherited Create(0);
  m_strWindowTitle    := 'TextDot3: Dot3 bumpmapped text (visit http://clootie.narod.ru)';
  m_bUseDepthBuffer   := False;
  m_dwCreationWidth   := 760;
  m_dwCreationHeight  := 265;

  // Create fonts
  m_strFont:= 'Arial';
  m_dwFontSize := 48;
  m_pFont      := CD3DFont.Create(m_strFont, m_dwFontSize);
  m_pDot3Font  := TD3DFontDot3.Create(m_strFont, m_dwFontSize);
  m_pStatsFont := CD3DFont.Create('Arial', 12, D3DFONT_BOLD);

  Hue:= 50; // initially display text in green
end;

//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
function CMyD3DApplication.OneTimeSceneInit: HRESULT;
begin
  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FrameMove: HRESULT;
var
  n: TD3DXVector3;
  pt: TPoint;
begin
  // Compute the light vector from the cursor position
  if GetFocus <> 0 then
  begin
    GetCursorPos(pt);
    ScreenToClient(m_hWnd, pt);

    n.x := (((2.0 * pt.x ) / m_d3dsdBackBuffer.Width ) - 1);
    n.y := (((2.0 * pt.y ) / m_d3dsdBackBuffer.Height ) - 1);
    n.z := 0.0;

    if (D3DXVec3Length(n) <= 1.0)
    then n.z := Sqrt(1.0 - n.x*n.x - n.y*n.y);

    MouseNormal:= n;
  end;

  // Calculate circular vector (for 3'rd text line)
  CircularNormal.x:= Cos(m_fTime*4);
  CircularNormal.y:= Sin(m_fTime*4);
  CircularNormal.z:= 1- (pt.y / m_d3dsdBackBuffer.Height);
  D3DXVec3Normalize(CircularNormal, CircularNormal);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.Render: HRESULT;
var
  rgb: TRGBRec;
begin
  // Clear the viewport
  m_pd3dDevice.Clear(0, nil, D3DCLEAR_TARGET, $00008800, 1.0, 0);

  // Begin the scene
  if SUCCEEDED(m_pd3dDevice.BeginScene) then
  begin
    // Draw CD3DFont in 2D (red)
    m_pFont.DrawText(20, 75, D3DCOLOR_ARGB(255,255,0,0), 'CD3DFont.DrawText');

    m_pDot3Font.SetLightNormal(MouseNormal);
    m_pDot3Font.DrawText(20, 135, D3DCOLOR_ARGB(255,0,0,255), 'TD3DFontDot3.DrawText');

    rgb:= HSBtoRGB(HSB(Hue, 240, 100));
    m_pDot3Font.SetLightNormal(CircularNormal);
    m_pDot3Font.DrawText(20, 190, D3DCOLOR_ARGB(255, rgb.r, rgb.g, rgb.b),
      'http://clootie.narod.ru');

    // Show frame rate
    m_pStatsFont.DrawText(2,  0, D3DCOLOR_ARGB(255,255,255,0), m_strFrameStats);

    // Show keyboard commands hints 
    m_pStatsFont.DrawText(5, 20, D3DCOLOR_ARGB(255,255,255,255),
      'F'#10 +
      'X'#10 +
      'A / Z');
    m_pStatsFont.DrawText(45, 20, D3DCOLOR_ARGB(255,255,255,0),
      '- Choose different font'#10 +
      '- Modulate 2X'#10 +
      '- Modify color hue of lowest text');

    m_pStatsFont.DrawText(305, 20, D3DCOLOR_ARGB(255,255,255,255),
      'D'#10 +
      'C');
    m_pStatsFont.DrawText(320, 20, D3DCOLOR_ARGB(255,255,255,0),
      '- Switch Bumped / "normal texture" view'#10 +
      '- Blend to Black / Back buffer');

    // End the scene.
    m_pd3dDevice.EndScene;
  end;

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.InitDeviceObjects: HRESULT;
begin
  // Restore the fonts
  m_pFont.InitDeviceObjects(m_pd3dDevice);
  m_pDot3Font.InitDeviceObjects(m_pd3dDevice);
  m_pStatsFont.InitDeviceObjects(m_pd3dDevice);

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.RestoreDeviceObjects: HRESULT;
var
  vEyePt: TD3DXVector3;
  vLookatPt: TD3DXVector3;
  vUpVec: TD3DXVector3;
  matWorld, matView, matProj: TD3DXMatrix;
  fAspect: Single;
begin
  // Restore the fonts
  m_pFont.RestoreDeviceObjects;
  m_pDot3Font.RestoreDeviceObjects;
  m_pStatsFont.RestoreDeviceObjects;

  // Restore the textures
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLOROP,   D3DTOP_MODULATE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_MINFILTER, D3DTEXF_LINEAR);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR);
  m_pd3dDevice.SetRenderState(D3DRS_DITHERENABLE, 1);
  m_pd3dDevice.SetRenderState(D3DRS_SPECULARENABLE, 1);
  m_pd3dDevice.SetRenderState(D3DRS_AMBIENT, $00ffffff);

  // Set the transform matrices
  vEyePt    := D3DXVector3(0.0,-5.0,-10.0);
  vLookatPt := D3DXVector3(0.0, 0.0,  0.0);
  vUpVec    := D3DXVector3(0.0, 1.0,  0.0);

  D3DXMatrixIdentity(matWorld);
  D3DXMatrixLookAtLH(matView, vEyePt, vLookatPt, vUpVec);
  fAspect := m_d3dsdBackBuffer.Width / m_d3dsdBackBuffer.Height;
  D3DXMatrixPerspectiveFovLH(matProj, D3DX_PI/4, fAspect, 1.0, 100.0);

  m_pd3dDevice.SetTransform(D3DTS_WORLD,      matWorld);
  m_pd3dDevice.SetTransform(D3DTS_VIEW,       matView);
  m_pd3dDevice.SetTransform(D3DTS_PROJECTION, matProj);

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.InvalidateDeviceObjects: HRESULT;
begin
  m_pFont.InvalidateDeviceObjects;
  m_pDot3Font.InvalidateDeviceObjects;
  m_pStatsFont.InvalidateDeviceObjects;

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.DeleteDeviceObjects: HRESULT;
begin
  m_pFont.DeleteDeviceObjects;
  m_pDot3Font.DeleteDeviceObjects;
  m_pStatsFont.DeleteDeviceObjects;

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FinalCleanup: HRESULT;
begin
  SAFE_DELETE(m_pFont);
  SAFE_DELETE(m_pDot3Font);
  SAFE_DELETE(m_pStatsFont);

  Result:= S_OK;
end;

procedure CMyD3DApplication.ChangeFont;
var
  hdc_: HDC;
  lHeight: Longint;
  lf: TLogFont;
  cf: TChooseFont;
begin
  hdc_ := GetDC(m_hWnd);
  lHeight := -MulDiv(m_dwFontSize, GetDeviceCaps(hdc_, LOGPIXELSY), 72);
  ReleaseDC(m_hWnd, hdc_);
  // hdc_ := 0;

  lstrcpy(lf.lfFaceName, m_strFont);
  lf.lfHeight := lHeight;

  ZeroMemory(@cf, SizeOf(cf));
  cf.lStructSize := SizeOf(cf);
  cf.hwndOwner   := m_hWnd;
  cf.lpLogFont   := @lf;
  cf.Flags       := CF_SCREENFONTS or CF_INITTOLOGFONTSTRUCT;

  if ChooseFont(cf) then
  begin
    SAFE_DELETE(m_pFont);
    SAFE_DELETE(m_pDot3Font);

    lstrcpy(m_strFont, lf.lfFaceName);
    m_dwFontSize := Trunc(cf.iPointSize / 10);

    m_pFont := CD3DFont.Create(m_strFont, m_dwFontSize);
    m_pFont.InitDeviceObjects(m_pd3dDevice);
    m_pFont.RestoreDeviceObjects;

    m_pDot3Font:= TD3DFontDot3.Create(m_strFont, m_dwFontSize);
    m_pDot3Font.InitDeviceObjects(m_pd3dDevice);
    m_pDot3Font.RestoreDeviceObjects;
  end;
end;

//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: Overrrides the main WndProc, so the sample can do custom message
//       handling (e.g. processing mouse, keyboard, or menu commands).
//-----------------------------------------------------------------------------
function CMyD3DApplication.MsgProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM;
  lParam: LPARAM): LRESULT;
begin
  // Handle key presses
  if (WM_KEYDOWN = uMsg) then
  begin
    if (Ord('D') = wParam) then
      m_pDot3Font.fOnlyDot3:= not m_pDot3Font.fOnlyDot3;

    if (Ord('X') = wParam) then
      m_pDot3Font.fModulate2X:= not m_pDot3Font.fModulate2X;

    if (Ord('C') = wParam) then
      m_pDot3Font.fBlendToBlack:= not m_pDot3Font.fBlendToBlack;

    if (Ord('F') = wParam) then ChangeFont;

    if (Ord('A') = wParam) then
    begin
      Inc(Hue, 5);
      if Hue > 235 then Hue:= 0;
    end;
    if (Ord('Z') = wParam) then
    begin
      if Hue > 5 then Dec(Hue, 5) else Hue:= 235;
    end;
  end;

  Result:= inherited MsgProc(hWnd, uMsg, wParam, lParam);
end;

end.

